# -*- coding: utf8 -*-
# This code is part of the opsi.org project
# https://www.opsi.org
#
# Copyright (c) 2024 uib GmbH <info@uib.de>
# This code is owned by the uib GmbH, Mainz, Germany (uib.de). All rights reserved.
# License: AGPL-3.0

import glob
import re
import shutil
from pathlib import Path

from opsibootimage import patcha
from opsisystem.bcd import BCD
from opsisystem.inffile import Architecture, INFFile, INFTargetOSVersion
from opsisystem.storage import get_disks


def update_product_on_clients() -> None:
	delete_pocs = []
	setup_pocs = []
	for poc in backend.productOnClient_getObjects(clientId = clientId):
		if poc.productType == 'NetbootProduct' and poc.productId != productId:
			delete_pocs.append(poc)
		elif poc.productType == 'LocalbootProduct':
			if poc.installationStatus == 'installed' and poc.productId not in ("opsi-client-agent", "opsi-script", "opsi-winst"):
				poc.actionRequest = "setup"
				setup_pocs.append(poc)
			elif poc.installationStatus == 'not_installed' and poc.actionRequest == 'none':
				delete_pocs.append(poc)
	if delete_pocs:
		backend.productOnClient_deleteObjects(delete_pocs)
	if setup_pocs:
		backend.productOnClient_updateObjectsWithDependencies(setup_pocs)


def set_action_progress(action_progress: str) -> None:
	backend.productOnClient_updateObject({
		"type": "ProductOnClient",
		"productType": "NetbootProduct",
		"productId": productId,
		"clientId": clientId,
		"actionProgress": action_progress
	})


def check_requirements() -> None:
	if not bootimageVersion or (int(bootimageVersion) < 20240228):
		raise RuntimeError("This package requires opsi-linux-bootimage version >= 20240228")


def run_hardware_audit() -> None:
	logger.notice("Fetching opsi hw audit configuration")
	hw_config = backend.auditHardware_getConfig()

	logger.notice("Running hardware inventory")
	audit_hardware_on_hosts = auditHardware(config=hw_config, hostId=clientId)

	logger.notice("Sending hardware information to service")
	backend.auditHardwareOnHost_setObsolete(clientId)
	backend.auditHardwareOnHost_createObjects(audit_hardware_on_hosts)
	context.audit_hardware_on_hosts = audit_hardware_on_hosts


def prepare_bios_disk() -> None:
	logger.notice("Swapoff to free devices")
	execute("swapoff -a -v")

	disks = sorted(get_disks(), key=lambda disk: disk.dev)
	logger.notice("Available disks: %s", disks)
	disk = disks[0]
	logger.notice("Using disk: %s", disk)
	partiton_table = disk.create_partition_table("MBR")
	context.boot_dev = partiton_table.create_partition(type="07", size="500MiB", bootable=True)
	context.win_dev = partiton_table.create_partition(type="07")
	context.boot_dev.create_filesystem("ntfs")
	context.win_dev.create_filesystem("ntfs")
	partiton_table.write_boot_record("mbr7")
	context.disk = disk
	context.boot_dev.set_mount_point("/mnt/boot")
	context.win_dev.set_mount_point("/mnt/windows")
	context.win_dev.set_windows_drive("c")


def prepare_uefi_disk() -> None:
	logger.notice("Swapoff to free devices")
	disks = sorted(get_disks(), key=lambda disk: disk.dev)
	logger.notice("Available disks: %s", disks)
	disk = disks[0]
	logger.notice("Using disk: %s", disk)
	disk.wipe_metadata()
	partiton_table = disk.create_partition_table("GPT")
	context.boot_dev = partiton_table.create_partition(number=1, type="c12a7328-f81f-11d2-ba4b-00a0c93ec93b", size="100MiB", name="EFI system partition")
	context.boot_dev.create_filesystem("vfat")
	windows_partition_number = 2
	if productProperties.get("setup_winre", True):
		# The Windows 10 recovery partition requires at least 522254916 bytes of free disk space.
		context.msr_dev = partiton_table.create_partition(number=2, type="de94bba4-06d1-4d40-a16a-bfd50179d6ac", size="600MiB", name="Recovery")
		context.msr_dev.create_filesystem("ntfs")
		windows_partition_number = 3
	context.win_dev = partiton_table.create_partition(number=windows_partition_number, type="ebd0a0a2-b9e5-4433-87c0-68b6b72699c7", name="WINDOWS")
	context.win_dev.create_filesystem("ntfs")
	context.disk = disk
	context.boot_dev.set_mount_point("/mnt/boot")
	context.win_dev.set_mount_point("/mnt/windows")
	context.win_dev.set_windows_drive("c")


def apply_wim(device: str) -> None:
	image = productPropertyValues.get("image", [""])[0]
	image_file, image_name = image.split(":", 1) if ":" in image else ("install.wim", image)
	image_path = context.script_path / "images" / image_file
	try:
		image_index = int(image_name)
		image_name = ""
		logger.info("Searching for image index %d in '%s'", image_name, image_path)
	except ValueError:
		image_index = -1
		logger.info("Searching for image %r in '%s'", image_name, image_path)

	cur_index = -1
	cur_name = ""
	cur_arch = ""
	for line in execute(f"wiminfo '{image_path}'"):
		line = line.strip()
		if line.startswith("Index:"):
			cur_index = int(line.split(":", 1)[-1].strip())
		elif line.startswith("Name:"):
			cur_name = line.split(":", 1)[-1].strip()
			logger.info("Image %r found at index %d", cur_name, cur_index)
		elif line.startswith("Architecture:"):
			cur_arch = line.split(":", 1)[-1].strip()
			logger.info("Architecture %r found at index %d", cur_arch, cur_index)
		elif not line and cur_index > -1 and ((image_name and image_name == cur_name) or (image_index > -1 and image_index == cur_index)):
			logger.notice("Using image %r at index %d", cur_name, cur_index)
			image_index = cur_index
			image_name = cur_name
			context.architecture = Architecture.from_string(cur_arch)
			break

	if image_index == -1:
		raise ValueError(f"Image '{image_name}' not found in '{image_path}'")
	if not image_name:
		raise ValueError(f"Image index '{image_index}' not found in '{image_path}'")
	if context.architecture == Architecture.X86 and context.boot_mode == "uefi":
		raise RuntimeError("32 bit Windows cannot be installed in UEFI mode")

	scriptMessageSubject.setMessage(
		f"apply {str(image_path)} with image: '{str(image_name.strip())}' to {device}"
	)
	logger.notice("Apply image %r (%d: %s) to %r", image_path, image_index, image_name, device)
	execute(f"wimapply '{image_path}' {image_index} '{device}'")


def integrate_drivers() -> None:
	# Integrate drivers
	driver_src = str(context.depot_script_path / "drivers")
	driver_dst = context.win_dev.abs_path(f"{context.opsi_org_dir}/drv")
	driver_dst.mkdir(parents=True, exist_ok=True)
	if not depot.exists(driver_src):
		logger.warning("Driver directory '%s' not found", driver_src)
		return

	logger.notice("Installing drivers from %s", driver_src)
	additional_drivers = [
		ad.strip() for ad in (','.join(productPropertyValues.get('additional_drivers', ['']))).split(',')
	]
	integrateAdditionalWindowsDrivers(
		driverSourceDirectory=f'{driver_src}/drivers/additional',
		driverDestinationDirectory=str(driver_dst),
		additionalDrivers=additional_drivers,
		messageSubject=scriptMessageSubject,
		srcRepository=depot,
		auditHardwareOnHosts=context.audit_hardware_on_hosts
	)
	integrateWindowsHardwareDrivers(
		driverSourceDirectory=driver_src,
		driverDestinationDirectory=str(driver_dst),
		auditHardwares=context.audit_hardware_on_hosts,
		messageSubject=scriptMessageSubject,
		srcRepository=depot
	)

	registry = ""
	for file_path in driver_dst.glob("**/*.[iI][nN][fF]"):
		logger.info("Found inf file: %s", file_path)
		driver_path = file_path.parent
		inf_file = INFFile(file_path)
		inf_file.parse()
		if inf_file.version.Class != "SCSIAdapter":
			continue
		logger.info("Integrating driver for storage controller: %s", inf_file.inf_name)

		arch = "amd64"
		driver_store_path = (
			context.win_dev.abs_path("Windows/System32/DriverStore/FileRepository") /
			inf_file.get_driver_database_dir_name(arch=context.architecture)
		)
		logger.info("Copy driver '%s' into driver store '%s'", driver_path, driver_store_path)
		shutil.copytree(driver_path, driver_store_path)

		win_drivers_path = context.win_dev.abs_path("Windows/System32/drivers")
		for sys_file in driver_path.glob("*.[sS][yY][sS]"):
			logger.info("Copy sys file '%s' into windows drivers '%s'", sys_file, win_drivers_path)
			shutil.copy(sys_file, win_drivers_path)
		try:
			registry += inf_file.get_driver_database_reg(
				target_os_version=INFTargetOSVersion(Architecture=context.architecture),
				oem_inf_name=inf_file.inf_name
			)
			registry += inf_file.get_services_reg(
				target_os_version=INFTargetOSVersion(Architecture=context.architecture)
			)
		except Exception as err:
			logger.error(err, exc_info=True)

	if registry:
		logger.info("Adding to registry: %s", registry)
		reg_file = Path("/tmp/driver.reg")
		reg_file.write_text("Windows Registry Editor Version 5.00\r\n\r\n" + registry, encoding="utf-16")
		reg_system = context.win_dev.abs_path("Windows/System32/config/SYSTEM")
		execute(f"reged -C -I '{str(reg_system)}' 'HKEY_LOCAL_MACHINE\\SYSTEM' '{str(reg_file)}'", ignoreExitCode=[2])


def copy_opsi_client_agent() -> None:
	program_files = "Program Files (x86)" if context.architecture == "x64" else "Program Files"
	context.program_files_opsi_dir = f"{program_files}/opsi.org"
	client_agent_dir = context.win_dev.abs_path(context.client_agent_dir)
	program_files_opsi_dir = context.win_dev.abs_path(context.program_files_opsi_dir)
	if client_agent_dir.exists():
		shutil.rmtree(opsi_org_dir)
	if program_files_opsi_dir.exists():
		shutil.rmtree(program_files_opsi_dir)
	depot.copy("/opsi-client-agent/*", str(client_agent_dir))
	copy(f"{str(client_agent_dir)}/files/opsi/postinst.d/*", str(context.win_dev.abs_path(context.postinst_dir) / "specialize"))


def copy_postinst_files() -> None:
	postinst_win_dir = context.win_dev.win_path(context.postinst_dir)
	postinst_dir = context.win_dev.abs_path(context.postinst_dir)
	depot.copy(f"{str(context.depot_script_path)}/postinst/*", str(postinst_dir))
	if depot.exists(f"{str(context.depot_script_path)}/custom/postinst"):
		depot.copy(f"{str(context.depot_script_path)}/custom/postinst/*", str(postinst_dir))

	data = (postinst_dir / "postinst.opsiscript").read_text(encoding="utf-8")
	data = re.sub(r'^DefVar\s+\$ProductId\$\s*=\s*"[^"]*"', f'DefVar $ProductId$ = "{productId}"', data, flags=re.IGNORECASE | re.MULTILINE)
	logger.debug("Writing postinst.opsiscript: %s", data)
	(postinst_dir / "postinst.opsiscript").write_text(data, encoding="utf-8")

	program_files = "Program Files (x86)" if context.architecture == "x64" else "Program Files"
	oca_path = context.win_dev.abs_path(f"{program_files}/opsi.org/opsi-client-agent")
	oca_path.mkdir(parents=True, exist_ok=True)
	with open(oca_path / ".opsiclientd_restart", mode="w", encoding="windows-1252") as file:
		#file.write(f"run_opsi_script=windows-win-install,{str(postinst_win_dir)}\\postinst.opsiscript\r\n")
		file.write(f"run_opsi_script={str(postinst_win_dir)}\\postinst.opsiscript\r\n")


def create_unattend_xml() -> None:
	unattend_xml = context.win_dev.abs_path(context.unattend_xml)
	unattend_xml.parent.mkdir(parents=True, exist_ok=True)
	unattend_xml_src = str(context.depot_script_path / "custom/unattend.xml")
	if not depot.exists(unattend_xml_src):
		unattend_xml_src = str(context.depot_script_path / "unattend.xml")
	depot.copy(unattend_xml_src, str(unattend_xml))

	data = unattend_xml.read_text(encoding="utf-8")
	arch = "amd64" if context.architecture == Architecture.X64 else "x86"
	data = re.sub(r'processorArchitecture="[^"]*"', f'processorArchitecture="{arch}"', data)
	logger.debug("Writing unattend.xml: %s", data)
	unattend_xml.write_text(data, encoding="utf-8")


def patch_files() -> None:
	files = glob.glob(f"{str(context.win_dev.abs_path(context.postinst_dir))}/**/*", recursive=True)
	files = [f for f in files if os.path.isfile(f)]
	unattend_xml = context.win_dev.abs_path(context.unattend_xml)
	files.append(str(unattend_xml))

	administrator_username = "Administrator"
	administrator_password = productProperties.get("administrator_password", "") or "nt123"
	system_keyboard_layout = productProperties.get("system_keyboard_layout", "") or "US keyboard (0409:00000409)"
	if match := re.search(r"(\d{4}:\d{8})", system_keyboard_layout):
		system_keyboard_layout = match.group(1)

	with open(context.patcha_in, mode='a', encoding="utf-8") as file:
		file.write(f"administrator_username={administrator_username}\n")
		file.write(f"administrator_password={administrator_password}\n")
		file.write(f"system_keyboard_layout={system_keyboard_layout}\n")
		client_agent_dir = str(context.win_dev.win_path(context.client_agent_dir)).replace("\\", "\\\\")
		file.write(f"client_agent_dir={client_agent_dir}\n")
		file.write("specialize_command=powershell.exe -ExecutionPolicy Bypass -WindowStyle hidden -File c:\\\\opsi.org\\\\postinst\\\\postinst.ps1 specialize\n")

	if not softwareLicenseKey:
		execute(f"sed -i '/ProductKey/d' {context.win_dev.abs_path(context.unattend_xml)}")

	for file in files:
		patcha.patch_file(file, params_file=context.patcha_in, kernel_params=True)

	# Remove empty product key tag from unattend.xml
	data = unattend_xml.read_text(encoding="utf-8")
	data = data.replace("<ProductKey></ProductKey>", "")
	unattend_xml.write_text(data, encoding="utf-8")


def install_bios_boot_files() -> None:
	context.boot_dev.write_boot_record("ntfs")
	shutil.copy(context.win_dev.abs_path("Windows/Boot/PCAT/bootmgr"), context.boot_dev.abs_path("bootmgr"))
	context.boot_dev.abs_path("Boot").mkdir(parents=True, exist_ok=True)
	# Write BCD
	disk_id = context.boot_dev.partition_table.id
	bcd = BCD(context.boot_dev.abs_path("Boot/BCD"), create_from_template=True)
	bcd.update_device_info(device_type="partition", disk_id=disk_id, partition_offset=context.win_dev.start)
	bcd.update_device_info(device_type="partition", disk_id=disk_id, partition_offset=context.boot_dev.start, entries=["{bootmgr}"])
	default = bcd.get_default_boot_entry_guid()
	bcd.update_boot_entry(entry=default, path=r"\windows\system32\winload.exe", description="opsi windows start", locale="en-US", system_root=r"\windows")
	bcd.update_boot_entry(entry="{bootmgr}", path=r"\EFI\Microsoft\Boot\bootmgfw.efi")

def install_uefi_boot_files() -> None:
	context.boot_dev.abs_path("EFI/Microsoft/Boot").mkdir(parents=True, exist_ok=True)
	context.boot_dev.abs_path("EFI/Boot").mkdir(parents=True, exist_ok=True)
	shutil.copy(context.win_dev.abs_path("Windows/Boot/EFI/bootmgfw.efi"), context.boot_dev.abs_path("EFI/Boot/bootx64.efi"))
	shutil.copytree(context.win_dev.abs_path("Windows/Boot/EFI/"), context.boot_dev.abs_path("EFI/Microsoft/Boot/"), dirs_exist_ok=True)
	execute('/bin/efibootmgr -c --disk %s --part 1 -L "Windows Boot Manager" --loader \\\\EFI\\\\Microsoft\\\\Boot\\\\bootmgfw.efi' % (context.disk.dev))

	# Write BCD
	disk_id = context.boot_dev.partition_table.id
	part_3_id = context.win_dev.partition_table.get_partition(3).uuid
	bcd = BCD(context.boot_dev.abs_path("EFI/Microsoft/Boot/BCD"), create_from_template=True)
	logger.notice(f"disk_id is: {disk_id}")
	bcd.update_device_info(
		device_type="partition",
		disk_id=disk_id,
		partition_id=part_3_id,
	)
	default = bcd.get_default_boot_entry_guid()
	bcd.update_boot_entry(
		entry=default,
		path=r"\Windows\System32\winload.efi",
		description="opsi windows start",
		locale="en-US",
		system_root=r"\Windows",
	)
	logger.info("BCD")
	logger.info(bcd.get_boot_entries())

	context.boot_dev.abs_path("EFI/Microsoft/Recovery").mkdir(parents=True, exist_ok=True)
	BCD(context.boot_dev.abs_path("EFI/Microsoft/Recovery/BCD"), create_from_template=True)


context.opsi_org_dir = "opsi.org"
context.postinst_dir = "opsi.org/postinst"
context.client_agent_dir = "opsi.org/tmp/opsi-client-agent"
context.unattend_xml = "Windows/Panther/unattend.xml"
context.architecture = Architecture.X64

check_requirements()

show_message("Running hardware audit", progress=10)
run_hardware_audit()

show_message("Preparing disk", progress=20)
if context.boot_mode == "uefi":
    prepare_uefi_disk()
else:
    prepare_bios_disk()

show_message("Applying WIM", progress=30)
apply_wim(context.win_dev.dev)
update_product_on_clients()
context.boot_dev.mount()
context.win_dev.mount()

show_message("Integrating drivers", progress=80)
integrate_drivers()

show_message("Installing files", progress=90)
create_unattend_xml()
copy_opsi_client_agent()
copy_postinst_files()
patch_files()
if context.boot_mode == "uefi":
	install_uefi_boot_files()
else:
	install_bios_boot_files()

context.boot_dev.umount()
context.win_dev.umount()

set_action_progress("Rebooting to specialize pass")
reboot()
